import sys
from typing import List, Optional, Union

import cv2
import numpy as np
import skimage
import skimage.exposure
import skimage.feature
import skimage.morphology
import torch
from skimage.filters import threshold_otsu, threshold_triangle
from skimage.util import img_as_ubyte

from towbintools.deep_learning import augmentation, util
from towbintools.foundation import binary_image, image_handling


def edge_based_segmentation(
    image: np.ndarray,
    pixelsize: float,
    sigma_canny: float = 1,
) -> np.ndarray:
    """
    Python adaptation of the OG Matlab code for Sobel-based segmentation

    Parameters:
            image (np.ndarray): The input 2D grayscale image as a NumPy array.
            pixelsize (float): Pixel size to consider when removing small objects.
            backbone (str, optional): The library to use for edge detection. Options are 'skimage' or 'opencv'. Default is 'skimage'.
            sigma_canny (float, optional): Standard deviation for the Gaussian filter used in Canny edge detection. Default is 1.

    Returns:
            np.ndarray: The segmented image as a binary mask (NumPy array).

    Raises:
            ValueError: If the input image is not 2D.
    """

    if image.ndim > 2:
        raise ValueError("Image must be 2D.")

    # image_dtype_max_value = np.iinfo(image.dtype).max
    thresh_otsu = threshold_otsu(image)
    edges = skimage.feature.canny(
        image.copy(),
        sigma=sigma_canny,
        low_threshold=thresh_otsu / 5,
        high_threshold=thresh_otsu / 2.5,
    ).astype(np.uint8)
    # edges = skimage.feature.canny(image.copy(), sigma=sigma_canny, low_threshold=image_dtype_max_value * 0.05, high_threshold=image_dtype_max_value*0.1).astype(np.uint8)

    edges = skimage.morphology.remove_small_objects(
        edges.astype(bool), 3, connectivity=2
    ).astype(np.uint8)

    contours, _ = cv2.findContours(edges, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)

    if len(contours) > 20000:
        return np.zeros_like(image).astype(np.uint8)

    edges = binary_image.connect_endpoints(edges, max_distance=200)

    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (5, 5))
    edges = (cv2.morphologyEx(edges, cv2.MORPH_CLOSE, kernel) > 0).astype(int)

    mask = binary_image.fill_bright_holes(image, edges, 10).astype(np.uint8)
    mask = skimage.morphology.remove_small_objects(
        mask.astype(bool), 422.5 / (pixelsize**2), connectivity=2
    ).astype(np.uint8)

    mask = binary_image.fill_bright_holes(image, mask, 5).astype(np.uint8)

    contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    if len(contours) > 0:
        out = np.zeros_like(mask)
        for contour in contours:
            cv2.drawContours(out, [contour], -1, 1, 1)

        threshold = np.percentile(image[out > 0], 30)  # type: ignore

        final_mask = (image > threshold).astype(np.uint8)
        final_mask = skimage.morphology.remove_small_objects(
            final_mask.astype(bool), 422.5 / (pixelsize**2), connectivity=2
        ).astype(np.uint8)

    else:
        final_mask = np.zeros_like(image).astype(np.uint8)

    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (5, 5))
    final_mask = (cv2.morphologyEx(final_mask, cv2.MORPH_CLOSE, kernel) > 0).astype(int)

    final_mask = binary_image.fill_bright_holes(image, final_mask, 10).astype(np.uint8)
    return final_mask


def deep_learning_segmentation(
    image,
    model,
    device,
    tiler,
    RGB=True,
    activation=None,
    batch_size=-1,
):
    # Split the image into tiles
    if tiler is None:
        tiles = [image]
    else:
        try:
            tiles = tiler.split(image)
        except:
            print("Error splitting image into tiles. Image shape: ", image.shape)
            sys.exit(1)

    # Prepare tiles for model input
    if RGB:
        tiles = [augmentation.grayscale_to_rgb(tile) for tile in tiles]
    else:
        tiles = [torch.tensor(tile[np.newaxis, ...]).unsqueeze(0) for tile in tiles]

    tiles_batched = (
        util.divide_batch(torch.stack(tiles), batch_size)
        if batch_size > 0
        else [torch.stack(tiles)]
    )

    prediction_tiles = []
    for batch in tiles_batched:
        batch = batch.to(device)

        # Predict
        with torch.no_grad():
            prediction = model(batch)
            if activation == "sigmoid":
                prediction = torch.sigmoid(prediction)
            elif activation == "softmax":
                prediction = torch.softmax(prediction, dim=1)

        # Post-process predictions
        for pred_tile in prediction:
            pred_tile = pred_tile.cpu().numpy()
            pred_tile = np.moveaxis(pred_tile, 0, -1)
            prediction_tiles.append(pred_tile)

    # Merge tiles to get the complete prediction
    if tiler is None:
        pred = prediction_tiles[0]
    else:
        pred = tiler.merge(prediction_tiles)
    mask = (pred > 0.5).astype(np.uint8)

    return mask


def double_threshold_segmentation(image):
    # keep bins 2**8 even though our images are 2**16 because none of the images cover the whole dynamic range of 2**16. This will bin lower abundance signal pixels into fewer histogram bins
    mask = image > _custom_threshold_otsu(image, nbins=2**8)
    mask = cv2.medianBlur(img_as_ubyte(mask), 5)
    mask = skimage.morphology.remove_small_objects(
        mask.astype(bool), 20, connectivity=2
    )
    mask = img_as_ubyte(mask)
    return mask


def segment_image(
    image: Union[str, np.ndarray],
    method: str,
    channels: Optional[List[int]] = None,
    pixelsize: float = None,
    sigma_canny: float = 1,
    preprocessing_fn=None,
    model=None,
    device=None,
    tiler=None,
    RGB=False,
    activation=None,
    batch_size=-1,
    is_zstack=True,
) -> np.ndarray:
    """
    Segment an image using the specified method.

    Parameters:
            image (Union[str, np.ndarray]): Input image. If string, it's interpreted as the path to a TIFF file. If ndarray, it's the image data directly.
            method (str): Segmentation method to use. Currently supported: "edge_based".
            channels (List[int], optional): List of channel indices to keep if reading a multi-channel TIFF file. Default is empty, meaning all channels are kept.
            pixelsize (Optional[float], optional): Pixel size to consider for edge-based segmentation. Must be specified if method is "edge_based".
            edge_based_backbone (str, optional): The library to use for edge detection in "edge_based" method. Options are 'skimage' or 'opencv'. Default is 'skimage'.
            sigma_canny (float, optional): Standard deviation for the Gaussian filter used in Canny edge detection if method is "edge_based". Default is 1.

    Returns:
            np.ndarray: The segmented image as a binary mask (NumPy array).

    Raises:
            ValueError: If method is not recognized or if pixelsize is not specified when required.
    """
    if type(image) == str:
        image = image_handling.read_tiff_file(image, channels_to_keep=channels)

    if method == "edge_based":
        if pixelsize is None:
            raise ValueError("Pixelsize must be specified for edge-based segmentation.")

        image = image_handling.normalize_image(image, dest_dtype=np.uint16)
        segment_fn = lambda x: edge_based_segmentation(
            x, pixelsize, sigma_canny=sigma_canny
        )

    elif method == "deep_learning":
        if not model:
            raise ValueError("Model must be specified for deep learning segmentation.")
        if not device:
            raise ValueError("Device must be specified for deep learning segmentation.")

        if preprocessing_fn:
            image = preprocessing_fn(image=image)["image"]
        segment_fn = lambda x: deep_learning_segmentation(
            x,
            model,
            device,
            tiler,
            RGB=RGB,
            activation=activation,
            batch_size=batch_size,
        )
    elif method == "double_threshold":
        segment_fn = double_threshold_segmentation

    else:
        raise ValueError("Invalid segmentation method.")

    if is_zstack:
        mask = np.zeros(
            (image.shape[0], image.shape[-2], image.shape[-1]), dtype=np.uint8
        )
        for i, plane in enumerate(image):
            mask[i] = segment_fn(plane).squeeze()
        return mask

    return segment_fn(image)


def _mode_limited_mean(image, nbins=2**8, histogram=None):
    """
    Brocher, Jan. (2014). Qualitative and Quantitative Evaluation of Two New Histogram Limiting Binarization Algorithms.
    International Journal of Image Processing (IJIP). 8. 30-48.
    Section 2.2 - Mode Limited Mean (MoLiM)
    """
    # basic idea of the paper and MoLiM is that most thresholding algorithms work on histograms
    # if most of the image is 'background', then the histogram will be dominated by a single background peak, which will affect how the threshold is calculated
    # MoLiM is the mean of the image with some of the background already classified as background
    # this limited mean value can either itself be a threshold, or can be used as a starting point for more sophisticated thresholding algorithms
    if histogram is None:
        hist, bin_centers = skimage.exposure.histogram(image, nbins=nbins)
    else:
        hist, bin_centers = histogram

    mode = bin_centers[hist.argmax()]
    molim = np.mean(image[image > mode])
    return molim


def _custom_threshold_otsu(image, nbins=2**8):
    # for consistent behaviour with regards to dtype and nbins, explicitly convert image to float
    # this is because skimage.filters.threshold_... and skimage.exposure.histogram ignore nbins in case of integer dtype
    # additionally, I think skimage trims histogram values below image.min() and above image.max() since they are zero
    # so all in all, when trying to troubleshoot threshold algorithm behaviour, and make it consistent across different channels, rescale the image as float in range (0, 1)

    # save the image range and dtype to later convert threshold back to what is expected
    image_range = image.min(), image.max()
    image_dtype = image.dtype
    image = skimage.exposure.rescale_intensity(image, out_range=(0, 1))
    # calculate histogram once and reuse in later thresholding steps
    histogram = skimage.exposure.histogram(image, nbins=nbins)
    hist_vals, bin_centers = histogram
    # limited_mean lies somewhere in the histogram peak for background
    limited_mean = _mode_limited_mean(image, histogram=histogram)
    # threshold_triangle by-and-large finds the threshold for the bottom of background peak
    # so it's a good starting point for otsu threshold, which can struggle in the case when background is so large as a proportion of the image that the overall histogram appears to be unimodal
    thresh_lower_bound = image.min()
    i = 0
    while thresh_lower_bound < limited_mean:
        # >= to allow possibility of thresh_lower_bound == limited_mean
        thresh_lower_bound = threshold_triangle(
            image[image >= thresh_lower_bound], nbins=nbins
        )
        if i > 64:
            break
        i += 1

    # don't start right at lower bound because the histogram information below thresh_lower_bound may be useful in determining a good threshold in the last iteration
    # for example, when thresh_lower_bound is already a decent threshold, in which case the the very information used to calculate it would have been discarded
    thresh = image.min()
    i = 0
    while thresh < thresh_lower_bound:
        # crop histogram to only include values above the current threshold
        hist_vals = hist_vals[bin_centers > thresh]
        bin_centers = bin_centers[bin_centers > thresh]
        thresh = threshold_otsu(hist=(hist_vals, bin_centers))

        if i > 64:
            break
        i += 1

    # rescale threshold back to original image range and dtype since threshold is in (0, 1)
    # rather than do the math ourselves, let skimage do it as it also handles float quantization
    thresh = skimage.exposure.rescale_intensity(
        np.array(thresh).reshape(1, 1), in_range=(0, 1), out_range=image_range
    ).astype(image_dtype)[0, 0]

    return thresh



import numpy as np
from skimage.measure import regionprops, shannon_entropy
from scipy.stats import skew, kurtosis
from skimage.feature import graycomatrix, graycoprops
from skimage.util import img_as_ubyte
from skimage.measure import regionprops_table


def compute_worm_volume(
    straightened_worm_mask: np.ndarray,
    pixelsize: float,
) -> float:
    """
    Compute the volume of a straightened worm mask using its radius and pixel size.

    Parameters:
        straightened_worm_mask (np.ndarray): The straightened worm mask as a NumPy array.
        pixelsize (float): The size of a pixel, used for volume calculation.

    Returns:
        float: The computed volume of the worm.
    """

    worm_radius = np.sum(straightened_worm_mask, axis=0) / 2
    return np.sum(np.pi * (worm_radius**2)) * (pixelsize**3)


def compute_worm_area(
    straightened_worm_mask: np.ndarray,
    pixelsize: float,
) -> float:
    """
    Compute the area of a worm mask.

    Parameters:
        straightened_worm_mask (np.ndarray): The straightened worm mask as a NumPy array.
        pixelsize (float): The size of a pixel, used for area calculation.

    Returns:
        float: The computed are of the worm.
    """

    return np.sum(straightened_worm_mask) * (pixelsize**2)


def compute_worm_length(
    straightened_worm_mask: np.ndarray,
    pixelsize: float,
) -> float:
    """
    Compute the length of a straightened worm mask using the sum of its pixels along the 0-axis.

    Parameters:
        straightened_worm_mask (np.ndarray): The straightened worm mask as a NumPy array.
        pixelsize (float): The size of a pixel, used for length calculation.

    Returns:
        float: The computed length of the worm.
    """
    return np.sum(np.sum(straightened_worm_mask, axis=0) > 0) * pixelsize


def compute_worm_type_features(
    straightened_worm_mask: np.ndarray,
    pixelsize: float,
) -> list:
    """
    Compute a series of morphological features for a straightened worm mask including
    length, volume, volume per length, width measures, entropy, and region properties.

    Parameters:
        straightened_worm_mask (np.ndarray): The straightened worm mask as a NumPy array.
        pixelsize (float): The size of a pixel, used for various calculations.

    Returns:
        list: A list containing the computed features in the following order:
              [worm_length, worm_volume, volume_per_length, width_mean, width_std,
              width_cv, entropy_mask, eccentricity, solidity, permimeter]
    """
    # compute worm length
    worm_length = compute_worm_length(straightened_worm_mask, pixelsize=pixelsize)

    # compute worm volume
    worm_volume = compute_worm_volume(straightened_worm_mask, pixelsize=pixelsize)

    # compute worm width
    worm_widths = np.sum(straightened_worm_mask, axis=0) * pixelsize
    worm_widths = worm_widths[worm_widths > 0]

    width_std = np.std(worm_widths)
    width_mean = np.mean(worm_widths)
    width_cv = width_std / width_mean

    volume_per_length = worm_volume / worm_length

    # compute entropy
    entropy_mask = shannon_entropy(straightened_worm_mask)

    try:
        other_properties = regionprops(straightened_worm_mask.astype(np.uint8))[0]
        eccentricity = other_properties.eccentricity
        solidity = other_properties.solidity
        permimeter = other_properties.perimeter
    except IndexError:
        eccentricity = 0
        solidity = 0
        permimeter = 0

    return [
        worm_length,
        worm_volume,
        volume_per_length,
        width_mean,
        width_std,
        width_cv,
        entropy_mask,
        eccentricity,
        solidity,
        permimeter,
    ]

# Features for nuclei classification

def intensity_std(regionmask, intensity_image):
    """
    Compute the standard deviation of the intensity values within a region.
    
    Parameters:
        regionmask (np.ndarray): The mask of the region.
        intensity_image (np.ndarray): The intensity image.
        
    Returns:
        float: The standard deviation of the intensity values within the region."""
    return np.std(intensity_image[regionmask])

def intensity_skew(regionmask, intensity_image):
    """
    Compute the skewness of the intensity values within a region.

    Parameters:
        regionmask (np.ndarray): The mask of the region.
        intensity_image (np.ndarray): The intensity image.
    
    Returns:
        float: The skewness of the intensity values within the region."""
    
    return skew(intensity_image[regionmask])

def intensity_kurtosis(regionmask, intensity_image):
    """
    Compute the kurtosis of the intensity values within a region.

    Parameters:
        regionmask (np.ndarray): The mask of the region.
        intensity_image (np.ndarray): The intensity image.
    
    Returns:
        float: The kurtosis of the intensity values within the region."""
    
    return kurtosis(intensity_image[regionmask])

def compute_haralick_features(image):
    """
    Compute Haralick texture features for an image.

    Parameters:
        image (np.ndarray): The intensity image.

    Returns:
        list: A list of Haralick texture features.
    """

    # Calculate the Grey-Level Co-Occurrence Matrix
    glcm = graycomatrix(img_as_ubyte(image), distances=[1], angles=[0], levels=256, symmetric=True, normed=True)
    # Calculate properties
    contrast = graycoprops(glcm, 'contrast')[0, 0]
    dissimilarity = graycoprops(glcm, 'dissimilarity')[0, 0]
    homogeneity = graycoprops(glcm, 'homogeneity')[0, 0]
    energy = graycoprops(glcm, 'energy')[0, 0]
    correlation = graycoprops(glcm, 'correlation')[0, 0]
    
    return [contrast, dissimilarity, homogeneity, energy, correlation]


def compute_patch_features(regionmask, intensity_image, patch_size=64):
    """
    Compute a set of features for a patch around a region.

    Parameters:
        regionmask (np.ndarray): The mask of the region.
        intensity_image (np.ndarray): The intensity image.
        patch_size (int): The size of the patch.

    Returns:
        list: A list of features for the patch.
    """

    # Get the centroid of the region
    centroid = regionprops(regionmask.astype("uint8"))[0].centroid
    # Create a patch of the defined size around the centroid
    minr = int(centroid[0] - patch_size/2)
    maxr = int(centroid[0] + patch_size/2)
    minc = int(centroid[1] - patch_size/2)
    maxc = int(centroid[1] + patch_size/2)

    if minr < 0:
        minr = 0
        maxr = patch_size
    if minc < 0:
        minc = 0
        maxc = patch_size
    if maxr >= intensity_image.shape[0]:
        maxr = intensity_image.shape[0] - 1
        minr = maxr - patch_size
    if maxc >= intensity_image.shape[1]:
        maxc = intensity_image.shape[1] - 1
        minc = maxc - patch_size

    patch = intensity_image[minr:maxr, minc:maxc]

    patch_basic_intensity_features = [np.max(patch), np.min(patch), np.mean(patch), np.std(patch), skew(patch.ravel()), kurtosis(patch.ravel())]
    patch_texture_features = compute_haralick_features(patch)
    patch_advanced_intensity_features = [shannon_entropy(patch)]

    patch_features = patch_basic_intensity_features + patch_texture_features + patch_advanced_intensity_features

    return patch_features

def compute_base_label_features(mask_of_label, intensity_image, features, extra_properties):
    """
    Compute a set of features for a single label.
    
    Parameters:
        mask_of_label (np.ndarray): The mask of the label.
        intensity_image (np.ndarray): The intensity image.
        features (list): The list of features to compute.
        extra_properties (list): The list of extra properties to compute.
        
    Returns:
        list: A list of features for the label.
    """

    properties = regionprops_table(mask_of_label, intensity_image=intensity_image, properties=features, extra_properties=extra_properties)  
    feature_vector = []
    for feature in properties:
        feature_vector.append(properties[feature][0])
    return feature_vector

def get_context(current_label, mask_of_current_label, mask_of_labels, num_closest=5):
    """
    Returns the mask of the closest labels to the current label in a label image.

    Parameters:
        current_label (int): The label of the current region.
        mask_of_current_label (np.ndarray): The mask of the current region.
        mask_of_labels (np.ndarray): The mask of all regions.
        num_closest (int): The number of closest regions to return.

    Returns:
        np.ndarray: The mask of the closest regions.
    """

    mask_of_all_other_labels = mask_of_labels.copy()
    mask_of_all_other_labels[mask_of_all_other_labels == current_label] = 0
    mask_of_all_other_labels = mask_of_all_other_labels.astype("uint8")

    if num_closest == -1:
        return mask_of_all_other_labels
    else:
        centroid_current_label = regionprops(mask_of_current_label)[0].centroid
        centroid_other_labels = regionprops(mask_of_all_other_labels)

        # find the num_closest labels
        closest_labels = sorted(centroid_other_labels, key=lambda x: np.linalg.norm(np.array(x.centroid) - np.array(centroid_current_label)))[:num_closest] # type: ignore
        closest_labels = [x.label for x in closest_labels]
        binary_mask_of_closest_labels = np.isin(mask_of_labels, closest_labels).astype("uint8")
        mask_of_closest_labels = mask_of_labels.copy()
        mask_of_closest_labels[binary_mask_of_closest_labels == 0] = 0
        return mask_of_closest_labels.astype("uint8")


def get_context_features(mask_of_labels, intensity_image, features, extra_properties):
    """
    Compute a set of features for the context of a label.

    Parameters:
        mask_of_labels (np.ndarray): The mask of all regions.
        intensity_image (np.ndarray): The intensity image.
        features (list): The list of features to compute.
        extra_properties (list): The list of extra properties to compute.

    Returns:
        list: A list of aggregated features for the context of the label (mean and std of each desired feature).
    """

    properties = regionprops_table(mask_of_labels, intensity_image=intensity_image, properties=features, extra_properties=extra_properties)
    context_feature_vector = []
    for feature in properties:
        context_feature_vector.append(np.mean(properties[feature]))
        context_feature_vector.append(np.std(properties[feature]))
    return context_feature_vector


import numpy as np
import xgboost

from towbintools.foundation import image_handling, worm_features
from typing import Callable
from joblib import Parallel, delayed
from towbintools.foundation.image_handling import check_if_zstack
import pandas as pd


def classify_worm_type(
    straightened_mask: np.ndarray,
    pixelsize: float,
    classifier: xgboost.XGBClassifier,
    classes: list = ["worm", "egg", "error"],
) -> str:
    """
    Classify the type of worm based on extracted features using an XGBoost classifier.

    The function extracts features from the provided straightened mask, then utilizes
    the XGBoost classifier to predict the type of worm. The classifier is assumed
    to return class probabilities, which are then converted to a one-hot encoding
    format to derive the final prediction.

    Parameters:
        straightened_mask (np.ndarray): The straightened worm mask for feature extraction.
        pixelsize (float): The pixel size for feature computations.
        classifier (xgboost.XGBClassifier): The trained XGBoost classifier object.
        classes (list[str], optional): List of classes that can be predicted by the classifier.
                                       Default is ['worm', 'egg', 'error'].

    Returns:
        str: The predicted class of the worm ('worm', 'egg', or 'error').
    """
    worm_type_features = np.array(
        [worm_features.compute_worm_type_features(straightened_mask, pixelsize)]
    )
    type_prediction = classifier.predict_proba(worm_type_features).squeeze()
    # convert proba to one hot encoding
    pred_class = np.argmax(type_prediction)
    prediction = classes[pred_class]
    return prediction


def classify_image(
    image: np.ndarray,
    features_function: Callable,
    classifier: xgboost.XGBClassifier,
    classes: list,
    **kwargs,
):
    """
    Classify images based on extracted features using a provided classifier.

    Parameters:
        images (np.ndarray): An array of images or a single image for feature extraction.
        features_function (callable): A function to extract features from the images.
        classifier (xgboost.XGBClassifier): The trained classifier object.
        classes (list[str]): List of classes that can be predicted by the classifier.
        return_proba (bool, optional): If True, return class probabilities instead of labels.
        **kwargs: Additional keyword arguments to pass to the features_function.

    Returns:
        np.ndarray or str: The predicted class of the image(s) or class probabilities.
    """

    # feature extraction
    try:
        features = features_function(image, **kwargs)
    except Exception as e:
        raise Exception(f"Error extracting features from image. {e}")
    # classification
    try:
        prediction = classifier.predict_proba(features).squeeze()
    except Exception as e:
        raise Exception(f"Error predicting class of image. {e}")

    assert len(prediction) == len(
        classes
    ), f"Number of provided classes and predicted classes do not match. len(prediction) = {len(prediction)}, len(classes) = {len(classes)}"
    # convert proba to one hot encoding
    pred_class = np.argmax(prediction)
    prediction = classes[pred_class]
    return prediction

def compute_features_of_label(current_label, mask_plane, image_plane, all_features, extra_properties, intensity_features, extra_intensity_features, num_closest=None, patches=None):
    """
    Compute a set of features for a single label, including context features and patch features.

    Parameters:
        current_label (int): The label of the current region.
        mask_plane (np.ndarray): The mask of all regions.
        image_plane (np.ndarray): The intensity image.
        all_features (list): The list of features to compute.
        extra_properties (list): The list of extra properties to compute.
        intensity_features (list): The list of intensity features to compute.
        extra_intensity_features (list): The list of extra intensity features to compute.
        num_closest (int): The number of closest regions to consider.
        patches (list): The list of patch sizes to consider.

    Returns:
        list: A list of features for the label.
    """
    
    mask_of_current_label = (mask_plane == current_label).astype("uint8")
    # check if image_plane has multiple channels
    if len(image_plane.shape) == 3:
        # compute all the features on the first channel and then intensity features on the other ones
        feature_vector = worm_features.compute_base_label_features(mask_of_current_label, image_plane[0], all_features, extra_properties)
        for i in range(1, image_plane.shape[0]):
            other_channel_intensity_features = worm_features.compute_base_label_features(mask_of_current_label, image_plane[i], intensity_features, extra_intensity_features)
            feature_vector += other_channel_intensity_features
    else:
        feature_vector = worm_features.compute_base_label_features(mask_of_current_label, image_plane, all_features, extra_properties)

    if patches is not None:
        for patch_size in patches:
            if len(image_plane.shape) == 3:
                patch_features = worm_features.compute_patch_features(mask_of_current_label, image_plane[0], patch_size=patch_size)
                for i in range(1, image_plane.shape[0]):
                    patch_features += worm_features.compute_patch_features(mask_of_current_label, image_plane[i], patch_size=patch_size)
                feature_vector += patch_features
            else:
                patch_features = worm_features.compute_patch_features(mask_of_current_label, image_plane, patch_size=patch_size)
                feature_vector += patch_features

    if num_closest is not None:
        context = worm_features.get_context(current_label, mask_of_current_label, mask_plane, num_closest=num_closest)
        if len(image_plane.shape) == 3:
            context_features = worm_features.get_context_features(context, image_plane[0], all_features, extra_properties)
            for i in range(1, image_plane.shape[0]):
                context_features += worm_features.get_context_features(context, image_plane[i], intensity_features, extra_intensity_features)
            feature_vector += context_features
        else:
            context_features = worm_features.get_context_features(context, image_plane, all_features, extra_properties)
            feature_vector += context_features

    return feature_vector

def compute_features_of_plane(mask_plane, image_plane, all_features, extra_properties, intensity_features, extra_intensity_features, num_closest=None, patches=None, parallel=True, n_jobs=-1):
    """
    Compute a set of features for a single label, including context features and patch features for all labels in a plane.

    Parameters:
        mask_plane (np.ndarray): The mask of all regions.
        image_plane (np.ndarray): The intensity image.
        all_features (list): The list of features to compute.
        extra_properties (list): The list of extra properties to compute.
        intensity_features (list): The list of intensity features to compute.
        extra_intensity_features (list): The list of extra intensity features to compute.
        num_closest (int): The number of closest regions to consider.
        patches (list): The list of patch sizes to consider.
        parallel (bool): Whether to compute features in parallel.
        n_jobs (int): The number of jobs to run in parallel.

    Returns:
        list: A list of lists of features for all labels.
    """

    if parallel:
        features_of_all_labels = Parallel(n_jobs=n_jobs)(delayed(compute_features_of_label)(current_label, mask_plane, image_plane, all_features, extra_properties, intensity_features, extra_intensity_features, num_closest=num_closest, patches=patches) for current_label in np.unique(mask_plane)[1:])
    else:
        features_of_all_labels = [compute_features_of_label(current_label, mask_plane, image_plane, all_features, extra_properties, intensity_features, extra_intensity_features, num_closest=num_closest, patches=patches) for current_label in np.unique(mask_plane)[1:]]
    return features_of_all_labels
    
def classify_plane(mask_plane, image_plane, classifier, all_features, extra_properties, intensity_features, extra_intensity_features, num_closest=None, patches=None, parallel=True, n_jobs=-1, confidence_threshold=None):
    """
    Compute the features of all the labels in a plane and classify them using an XGBoost classifier.

    Parameters:
        mask_plane (np.ndarray): The mask of all regions.
        image_plane (np.ndarray): The intensity image.
        classifier (xgboost.XGBClassifier): The trained classifier object.
        all_features (list): The list of features to compute.
        extra_properties (list): The list of extra properties to compute.
        intensity_features (list): The list of intensity features to compute.
        extra_intensity_features (list): The list of extra intensity features to compute.
        num_closest (int): The number of closest regions to consider.
        patches (list): The list of patch sizes to consider.
        parallel (bool): Whether to compute features in parallel.
        n_jobs (int): The number of jobs to run in parallel.
        confidence_threshold (float): The confidence threshold for predictions to be considered valid.

    Returns:
        list: A list of predicted classes for all labels in the plane.
    """
    
    features = compute_features_of_plane(mask_plane, image_plane, all_features, extra_properties, intensity_features, extra_intensity_features, num_closest=num_closest, patches=patches, parallel=parallel, n_jobs=n_jobs)
    if len(features) == 0:
        return None
    predictions = classifier.predict_proba(features)
    predicted_classes = np.argmax(predictions, axis=1)
    if confidence_threshold is not None:
        for i in range(len(predicted_classes)):
            if np.max(predictions[i]) < confidence_threshold:
                predicted_classes[i] = -1
    return predicted_classes

def classify_labels(mask, image, classifier, all_features, extra_properties, intensity_features, extra_intensity_features, num_closest=None, patches=None, parallel=True, n_jobs=-1, is_zstack=False, confidence_threshold=None):
    """
    Compute the features of all the labels in a mask and classify them using an XGBoost classifier.

    Parameters:
        mask (np.ndarray): The mask of all regions.
        image (np.ndarray): The intensity image.
        classifier (xgboost.XGBClassifier): The trained classifier object.
        all_features (list): The list of features to compute.
        extra_properties (list): The list of extra properties to compute.
        intensity_features (list): The list of intensity features to compute.
        extra_intensity_features (list): The list of extra intensity features to compute.
        num_closest (int): The number of closest regions to consider.
        patches (list): The list of patch sizes to consider.
        parallel (bool): Whether to compute features in parallel.
        n_jobs (int): The number of jobs to run in parallel.
        is_zstack (bool): Whether the image is a z-stack.
        confidence_threshold (float): The confidence threshold for predictions to be considered valid.

    Returns:
        list: A list of predicted classes for all labels in the mask.
    """

    if is_zstack or len(image.shape) > 3:
        assert mask.shape[0] == image.shape[0], "The number of planes in the mask and the image should be the same."
        return [classify_plane(mask_plane, image_plane, classifier, all_features, extra_properties, intensity_features, extra_intensity_features, num_closest=num_closest, patches=patches, parallel=parallel, n_jobs=n_jobs, confidence_threshold=confidence_threshold) for mask_plane, image_plane in zip(mask, image)]
    else:
        return classify_plane(mask, image, classifier, all_features, extra_properties, intensity_features, extra_intensity_features, num_closest=num_closest, patches=patches, parallel=parallel, n_jobs=n_jobs, confidence_threshold=confidence_threshold)

def classify_labels_features_dict(mask, image, clf, features_dict, parallel=True, n_jobs=-1, is_zstack=False, confidence_threshold=None):
    return classify_labels(mask, image, clf, features_dict['all_features'], features_dict['extra_properties'], features_dict['intensity_features'], features_dict['extra_intensity_features'], num_closest=features_dict['num_closest'], patches=features_dict['patches'], parallel=parallel, n_jobs=n_jobs, is_zstack=is_zstack, confidence_threshold=confidence_threshold)

def convert_classification_to_mask(mask, classification, is_zstack=False):
    """
    Convert a classification (list of predicted classes) to a mask.

    Parameters:
        mask (np.ndarray): The mask of all regions.
        classification (list): The list of predicted classes for all labels.
        is_zstack (bool): Whether the image is a z-stack.

    Returns:
        np.ndarray: The given mask with pixel values replaced with class number + 1.
    """

    new_mask = np.zeros_like(mask)

    if is_zstack or len(mask.shape) > 2:    
        for i, plane_classification in enumerate(classification):
            if plane_classification is not None:
                for j, label in enumerate(np.unique(mask[i])[1:]):
                    new_mask[i][mask[i] == label] = plane_classification[j] + 1
    else:
        if classification is not None:
            for i, label in enumerate(np.unique(mask)[1:]):
                new_mask[mask == label] = classification[i] + 1

    return new_mask

def convert_classification_to_dataframe(mask, classification, is_zstack=False):
    """
    Convert a classification (list of predicted classes) to a pandas DataFrame.

    Parameters:
        mask (np.ndarray): The mask of all regions.
        classification (list): The list of predicted classes for all labels.
        is_zstack (bool): Whether the image is a z-stack.

    Returns:
        pd.DataFrame: A DataFrame with columns "Plane", "Label", and "Class".
    """

    data = []
    if is_zstack or len(mask.shape) > 2:
        for i, plane_classification in enumerate(classification):
            if plane_classification is not None:
                for j, label in enumerate(np.unique(mask[i])[1:]):
                    data.append({"Plane": i, "Label": int(label), "Class": plane_classification[j]})
    else:
        if classification is not None:
            for i, label in enumerate(np.unique(mask)[1:]):
                data.append({"Plane": 0, "Label": int(label), "Class": classification[i]})
    return pd.DataFrame(data)

def classify_labels_and_convert_to_mask(mask, image, classifier, all_features, extra_properties, intensity_features, extra_intensity_features, num_closest=None, patches=None, parallel=True, n_jobs=-1, is_zstack=False, confidence_threshold=None):
    """
    Classify all the labels in a mask using an XGBoost classifier and convert the classification to a mask.
    
    Parameters:
        mask (np.ndarray): The mask of all regions.
        image (np.ndarray): The intensity image.
        classifier (xgboost.XGBClassifier): The trained classifier object.
        all_features (list): The list of features to compute.
        extra_properties (list): The list of extra properties to compute.
        intensity_features (list): The list of intensity features to compute.
        extra_intensity_features (list): The list of extra intensity features to compute.
        num_closest (int): The number of closest regions to consider.
        patches (list): The list of patch sizes to consider.
        parallel (bool): Whether to compute features in parallel.
        n_jobs (int): The number of jobs to run in parallel.
        is_zstack (bool): Whether the image is a z-stack.
        confidence_threshold (float): The confidence threshold for predictions to be considered valid.

    Returns:
        np.ndarray: The given mask with pixel values replaced with class number + 1.
    """
    
    classification = classify_labels(mask, image, classifier, all_features, extra_properties, intensity_features, extra_intensity_features, num_closest=num_closest, patches=patches, parallel=parallel, n_jobs=n_jobs, is_zstack=is_zstack, confidence_threshold=confidence_threshold)
    return convert_classification_to_mask(mask, classification)

def classify_labels_and_convert_to_dataframe(mask, image, classifier, all_features, extra_properties, intensity_features, extra_intensity_features, num_closest=None, patches=None, parallel=True, n_jobs=-1, is_zstack=False, confidence_threshold=None):
    """
    Classify all the labels in a mask using an XGBoost classifier and convert the classification to a pandas DataFrame.

    Parameters:
        mask (np.ndarray): The mask of all regions.
        image (np.ndarray): The intensity image.
        classifier (xgboost.XGBClassifier): The trained classifier object.
        all_features (list): The list of features to compute.
        extra_properties (list): The list of extra properties to compute.
        intensity_features (list): The list of intensity features to compute.
        extra_intensity_features (list): The list of extra intensity features to compute.
        num_closest (int): The number of closest regions to consider.
        patches (list): The list of patch sizes to consider.
        parallel (bool): Whether to compute features in parallel.
        n_jobs (int): The number of jobs to run in parallel.
        is_zstack (bool): Whether the image is a z-stack.
        confidence_threshold (float): The confidence threshold for predictions to be considered valid.

    Returns:
        pd.DataFrame: A DataFrame with columns "Plane", "Label", and "Class".
    """
    
    classification = classify_labels(mask, image, classifier, all_features, extra_properties, intensity_features, extra_intensity_features, num_closest=num_closest, patches=patches, parallel=parallel, n_jobs=n_jobs, is_zstack=is_zstack, confidence_threshold=confidence_threshold)
    return convert_classification_to_dataframe(mask, classification)

def classify_labels_and_convert_to_mask_features_dict(mask, image, clf, features_dict, parallel=True, n_jobs=-1, is_zstack=False, confidence_threshold=None):
    classification = classify_labels_features_dict(mask, image, clf, features_dict, parallel=parallel, n_jobs=n_jobs, is_zstack=is_zstack, confidence_threshold=confidence_threshold)
    return convert_classification_to_mask(mask, classification, is_zstack=is_zstack)

def classify_labels_and_convert_to_dataframe_features_dict(mask, image, clf, features_dict, parallel=True, n_jobs=-1, is_zstack=False, confidence_threshold=None):
    classification = classify_labels_features_dict(mask, image, clf, features_dict, parallel=parallel, n_jobs=n_jobs, is_zstack=is_zstack, confidence_threshold=confidence_threshold)
    return convert_classification_to_dataframe(mask, classification, is_zstack=is_zstack)



import numpy as np
import pandas as pd
from scipy.signal import savgol_filter, medfilt
from towbintools.foundation.utils import interpolate_nans
from scipy.ndimage import uniform_filter1d

def compute_growth_rate_linear(series, time, ignore_start_fraction=0., ignore_end_fraction=0., savgol_filter_window=5, savgol_filter_order=3):
    """
    Compute the growth rate of a time series using linear regression.
    """

    # Assert that the series and time have the same length
    assert len(series) == len(time), "The series and time must have the same length."

    # Compute the number of points to ignore at the beginning and end
    num_points = len(series)
    num_ignore_start = int(ignore_start_fraction * num_points)
    num_ignore_end = int(ignore_end_fraction * num_points)

    # Assert that the fraction of points to ignore is not too large
    assert num_ignore_start + num_ignore_end < num_points, "The fraction of points to ignore is too large."

    # Correctly handle slicing when ignore fractions are 0
    if num_ignore_end == 0:
        time = time[num_ignore_start:]
        series = series[num_ignore_start:]
    else:
        time = time[num_ignore_start:-num_ignore_end]
        series = series[num_ignore_start:-num_ignore_end]
    
    # Remove extreme outliers with a small median filter
    series = medfilt(series, 3)

    # Smooth the series time series a bit more with a Savitzky-Golay filter
    series = savgol_filter(series, savgol_filter_window, savgol_filter_order)

    # Compute linear regression
    slope, intercept = np.polyfit(time, series, 1)

    return slope

def compute_growth_rate_exponential(series, time, ignore_start_fraction=0., ignore_end_fraction=0., savgol_filter_window=5, savgol_filter_order=3):
    """
    Compute the growth rate of a time series using exponential regression.
    """

    # Assert that the series and time have the same length
    assert len(series) == len(time), "The series and time must have the same length."

    # Compute the number of points to ignore at the beginning and end
    num_points = len(series)
    num_ignore_start = int(ignore_start_fraction * num_points)
    num_ignore_end = int(ignore_end_fraction * num_points)

    # Assert that the fraction of points to ignore is not too large
    assert num_ignore_start + num_ignore_end < num_points, "The fraction of points to ignore is too large."

    # Correctly handle slicing when ignore fractions are 0
    if num_ignore_end == 0:
        time = time[num_ignore_start:]
        series = series[num_ignore_start:]
    else:
        time = time[num_ignore_start:-num_ignore_end]
        series = series[num_ignore_start:-num_ignore_end]
    
    # Remove extreme outliers with a small median filter
    series = medfilt(series, 3)

    # Smooth the series time series a bit more with a Savitzky-Golay filter
    series = savgol_filter(series, savgol_filter_window, savgol_filter_order)

    # Compute exponential regression
    slope, intercept = np.polyfit(time, np.log(series), 1)

    return slope

def compute_instantaneous_growth_rate(series, time, smoothing_method = "savgol", savgol_filter_window=15, savgol_filter_order=3, moving_average_window=15):
    """
    Compute the instantaneous growth rate of a time series.
    """

    # Assert that the series and time have the same length
    assert len(series) == len(time), "The series and time must have the same length."
    
    # Remove extreme outliers with a small median filter
    series = medfilt(series, 3)

    if smoothing_method == "savgol":
        # Smooth the series time series a bit more with a Savitzky-Golay filter
        series = savgol_filter(series, savgol_filter_window, savgol_filter_order)
    elif smoothing_method == "moving_average":
        # Smooth the series time series a bit more with a moving average filter
        series = uniform_filter1d(series, size=moving_average_window)

    # Compute the instantaneous growth rate
    growth_rate = np.gradient(series, time)

    return growth_rate

def compute_growth_rate_classified(series, time, worm_type, method='exponential', ignore_start_fraction=0., ignore_end_fraction=0., savgol_filter_window=5, savgol_filter_order=3):
    """
    Compute the growth rate of a time series, using only points correctly classified as worms.
    """

    # Assert that the series, time, and worm_type have the same length
    assert len(series) == len(time) == len(worm_type), "The series, time, and worm_type must have the same length."

    # Correct the series time series
    series_worms = correct_series_time_series(series, worm_type)

    if method == 'exponential':
        growth_rate = compute_growth_rate_exponential(series_worms, time, ignore_start_fraction, ignore_end_fraction, savgol_filter_window, savgol_filter_order)
    elif method == 'linear':
        growth_rate = compute_growth_rate_linear(series_worms, time, ignore_start_fraction, ignore_end_fraction, savgol_filter_window, savgol_filter_order)
    
    return growth_rate

def compute_instantaneous_growth_rate_classified(series, time, worm_type, smoothing_method = "savgol", savgol_filter_window=15, savgol_filter_order=3, moving_average_window=15):
    """
    Compute the instantaneous growth rate of a time series, using only points correctly classified as worms.
    """

    # Assert that the series, time, and worm_type have the same length
    assert len(series) == len(time) == len(worm_type), "The series, time, and worm_type must have the same length."

    # Correct the series time series
    series_worms = correct_series_time_series(series, worm_type)
    growth_rate = compute_instantaneous_growth_rate(series_worms, time, smoothing_method, savgol_filter_window, savgol_filter_order, moving_average_window)
    
    return growth_rate

def correct_series_time_series(series, worm_type):
    """
    Remove the series of non-worms from the time series and interpolate them back.
    """

    # Set the series of non worms to NaN
    non_worms_indices = worm_type != 'worm'
    series_worms = series.copy()
    series_worms[non_worms_indices] = np.nan

    # Interpolate the NaNs
    series_worms = interpolate_nans(series_worms)
    
    return series_worms

def compute_growth_rate_per_larval_stage(series, time, worm_type, ecdysis, method = "exponential", ignore_start_fraction=0., ignore_end_fraction=0., savgol_filter_window=5, savgol_filter_order=3):
    """
    Compute the growth rate of a time series per larval stage.
    """

    # Assert that the series, time, and worm_type have the same length
    assert len(series) == len(time) == len(worm_type), "The series, time, and worm_type, must have the same length."

    # Correct the series time series
    series_worms = correct_series_time_series(series, worm_type)

    # extract ecdisis indices
    hatch_time = ecdysis['HatchTime']
    M1 = ecdysis['M1']
    M2 = ecdysis['M2']
    M3 = ecdysis['M3']
    M4 = ecdysis['M4']

    growth_rates = {}

    # Compute the growth rate per larval stage
    for i, (start, end) in enumerate(zip([hatch_time, M1, M2, M3], [M1, M2, M3, M4])):
        # check if start or end is NaN
        if np.isnan(start) or np.isnan(end):
            growth_rates[f"L{i+1}"] = np.nan
            
        else:
            series_worms_stage = series_worms[start:end]
            time_stage = time[start:end]
            worm_type_stage = worm_type[start:end]

            growth_rate_stage = compute_growth_rate_classified(series_worms_stage, time_stage, worm_type_stage, method, ignore_start_fraction, ignore_end_fraction, savgol_filter_window, savgol_filter_order)
            growth_rates[f"L{i+1}"] = growth_rate_stage
    
    return growth_rates